@@ -83,7 +83,7 @@ def geo_download(geo_id, series_path, geo_platforms, clean=True, decompress=True
 
     for platform in geo_platforms:
         if not Path(f"{series_path}/{platform}").exists():
-            Path(f"{series_path}/{platform}").mkdir()
+            Path(f"{series_path}/{platform}").mkdir(parents=True, exist_ok=True)
 
     ftp = FTP('ftp.ncbi.nlm.nih.gov', timeout=120) # 2 mins
     ftp.login()
@@ -91,12 +91,26 @@ def geo_download(geo_id, series_path, geo_platforms, clean=True, decompress=True
 
     try:
         filesize = ftp.size(f"miniml/{miniml_filename}.tgz") # -- gives 550 error because CWD puts it in ASCII mode.
-        #LOGGER.info(f"DEBUG ftp.size WORKED: miniml/{miniml_filename}.tgz -- {filesize}")
     except Exception as e:
         LOGGER.error(f"ftp.size ERROR: {e}")
 
+    def is_valid_tgz(filepath=f"{series_path}/{miniml_filename}.tgz"):
+        if Path(filepath).exists():
+            try:
+                min_tar = tarfile.open(filepath)
+                for file in min_tar.getnames():
+                    pass
+                return True
+            except gzip.BadGzipFile:
+                LOGGER.info(f"Found a family series file, but appears to be corrupted, so re-downloading.")
+                Path(filepath).unlink()
+                return False
+        else:
+            return False # file does not exist yet, so download
+
     if not Path(f"{series_path}/{miniml_filename}").exists():
-        if not Path(f"{series_path}/{miniml_filename}.tgz").exists():
+        # test opening it; may be corrupted
+        while is_valid_tgz(f"{series_path}/{miniml_filename}.tgz") is False:
             LOGGER.info(f"Downloading {miniml_filename}")
             miniml_file = open(f"{series_path}/{miniml_filename}.tgz", 'wb')
             try:
@@ -109,15 +123,14 @@ def geo_download(geo_id, series_path, geo_platforms, clean=True, decompress=True
                     def tqdm_callback(data):
                         tqdm_instance.update(len(data))
                         miniml_file.write(data)
-                    ftp.retrbinary(f"RETR miniml/{miniml_filename}.tgz", tqdm_callback)
+                    ftp.retrbinary(f"RETR miniml/{miniml_filename}.tgz", tqdm_callback, 1024)
             except Exception as e:
                 LOGGER.error(e)
-                LOGGER.info('tqdm: Failed to create a progress bar, but it is downloading...')
-                ftp.retrbinary(f"RETR miniml/{miniml_filename}.tgz", miniml_file.write)
+                LOGGER.info('tqdm: Failed to create a progress bar, but it should be downloading...')
+                ftp.retrbinary(f"RETR miniml/{miniml_filename}.tgz", miniml_file.write, 1024)
             miniml_file.close()
-            #LOGGER.info(f"Downloaded {miniml_filename}")
+
         #ftp.quit() # instead of 'close()'
-        #LOGGER.info(f"Unpacking {miniml_filename}")
         min_tar = tarfile.open(f"{series_path}/{miniml_filename}.tgz")
         for file in min_tar.getnames():
             if file == miniml_filename:
@@ -136,22 +149,35 @@ def geo_download(geo_id, series_path, geo_platforms, clean=True, decompress=True
             raw_file = open(f"{series_path}/{raw_filename}", 'wb')
             filesize = ftp.size(f"suppl/{raw_filename}")
             try:
-                try:
-                    with tqdm(unit = 'b', unit_scale = True, leave = False, miniters = 1, desc = geo_id, total = filesize) as tqdm_instance:
-                        def tqdm_callback(data):
-                            tqdm_instance.update(len(data))
-                            raw_file.write(data)
-                        ftp.retrbinary(f"RETR suppl/{raw_filename}", tqdm_callback)
-                except Exception as e:
-                    LOGGER.info('tqdm: Failed to create a progress bar, but it is downloading...')
-                    ftp.retrbinary(f"RETR suppl/{raw_filename}", raw_file.write)
+                with tqdm(unit = 'b', unit_scale = True, leave = False, miniters = 1, desc = geo_id, total = filesize) as tqdm_instance:
+                    def tqdm_callback(data):
+                        # update progress bar and write out data
+                        tqdm_instance.update(len(data))
+                        raw_file.write(data)
+
+                        # check to see if download has completed
+                        if (raw_file.tell() == filesize):
+                            # close down the progress bar
+                            # prevents an extra ftp call which generates a timed out exception
+                            tqdm_instance.close()
+
+                    # make ftp data request
+                    ftp.retrbinary(f"RETR suppl/{raw_filename}", tqdm_callback)
+
+                # quit the current ftp session
                 ftp.quit()
-            except socket.timeout as e:
-                LOGGER.warning(f"FTP timeout error.")
-                # seems to happen AFTER download is done, so just ignoring it.
-            LOGGER.info(f"Closing file {raw_filename}")
+
+            except Exception as e:
+                LOGGER.warning(e)
+
+            # check the downloaded file size
+            rawsize = raw_file.tell()
+            if rawsize != filesize:
+                LOGGER.info('geo_download sizes: FTP:{0} Raw:{1} Diff:{2}'.format(filesize, rawsize, filesize-rawsize))
+                LOGGER.info(f"geo_download: File {raw_filename} download failed.")
+                return False
             raw_file.close()
-            #LOGGER.info(f"Downloaded {raw_filename}")
+
         LOGGER.info(f"Unpacking {raw_filename}")
         try:
             tar = tarfile.open(f"{series_path}/{raw_filename}")
@@ -171,22 +197,17 @@ def geo_download(geo_id, series_path, geo_platforms, clean=True, decompress=True
         if clean:
             os.remove(f"{series_path}/{raw_filename}")
 
-    if not decompress:
-        pass #LOGGER.info(f"Not decompressing {geo_id} IDAT files")
-    else:
-        #LOGGER.info(f"Decompressing {geo_id} IDAT files")
+    if decompress:
         for gz in series_dir.glob("*.idat.gz"):
             gz_string = str(gz)
             with gzip.open(gz_string, 'rb') as f_in:
                 with open(gz_string[:-3], 'wb') as f_out:
                     shutil.copyfileobj(f_in, f_out)
             if clean:
-                gz.unlink() #os.remove(gz_string)
-
-    if not decompress:
-        LOGGER.info(f"Downloaded {geo_id} idats without decompressing")
-    else:
+                gz.unlink()
         LOGGER.info(f"Downloaded and unpacked {geo_id} idats")
+    elif not decompress:
+        LOGGER.info(f"Downloaded {geo_id} idats without decompressing")
     return success
 
 
@@ -362,7 +383,7 @@ NOTE: v1.3.0 does NOT support multiple GEO IDs yet.
         working = tempfile.TemporaryDirectory(dir=EFS)
     elif not kwargs.get('efs') and kwargs.get('data_dir') != None:
         if not Path(kwargs.get('data_dir')).exists():
-            Path(kwargs.get('data_dir')).mkdir()
+            Path(kwargs.get('data_dir')).mkdir(parents=True, exist_ok=True)
         working = tempfile.TemporaryDirectory(dir=kwargs['data_dir'])
         EFS = working.name
     elif not kwargs.get('efs') and kwargs.get('data_dir') == None:
