@@ -1,11 +1,10 @@
-from .pipeline import SampleDataContainer, get_manifest, run_pipeline, make_pipeline
+from .pipeline import SampleDataContainer, run_pipeline, make_pipeline
 from .preprocess import preprocess_noob
 from .postprocess import consolidate_values_for_sheet
 from .read_geo_processed import read_geo, detect_header_pattern
 
 __all__ = [
     'SampleDataContainer',
-    'get_manifest',
     'preprocess_noob',
     'run_pipeline',
     'make_pipeline,'
